/* Tunguska, ternary virtual machine
 *
 * Copyright (C) 2007,2008 Viktor Lofgren
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */

#include <stdlib.h>
#include <stdio.h>
#include <map>
#include <string>
#include <getopt.h>
#include <iostream>
#include <fstream>
#include <sstream>
#include <FlexLexer.h>
#include "machine.h"
#include "agdp.h"
#include "assembler.h"
#include "error.h"

using namespace std;

char* filename;
extern int yyparse();

/* C style loose assembler helper functions
 * 
 *
 */

/* see share/symbols.bmp */
char asciitoternary(char a) {
	char translation[] = {  0, ' ','\n','\t','\b',   0,  0,   0,   0,   0,
			    'A', 'B', 'C', 'D', 'E', 'F', 'G', 'H', 'I', 'J',
			    'K', 'L', 'M', 'N', 'O', 'P', 'Q', 'R', 'S', 'T',
			    'U', 'V', 'W', 'X', 'Y', 'Z', '0', '1', '2', '3',
			    '4', '5', '6', '7', '8', '9', '.', ',', '!', '?',
			    'a', 'b', 'c', 'd', 'e', 'f', 'g', 'h', 'i', 'j',
			    'k', 'l', 'm', 'n', 'o', 'p', 'q', 'r', 's', 't',
			    'u', 'v', 'w', 'x', 'y', 'z',  0 ,   0,   0,   0,
			      0,   0,   0,   0,   0,   0, '=', '-', '*', '/',
			    '%', '<', '>',   0,   0, '(', ')', '$', '+', '#' };

	for(int c  = 0; c < 100; c++) {
		if(translation[c] == a) return c;

	}

	return 0;
}

int lowtryte(int v) {
	tryte low, high;
	tryte::int_to_word(v, high, low);
	return low.to_int();
}
int hightryte(int v) {
	tryte low, high;
	tryte::int_to_word(v, high, low);
	return high.to_int();
}

int nontriplet(char* s) {
	tryte t = (s+1);
	return t.to_int();
}

int nonsextet(char* s) {
	if(strlen(s) != 7) throw new error(WHERE, "Bad nonary sextet");
	char s1[4];
	char s2[4];
	s1[0] = s[1]; s1[1] = s[2]; s1[2] = s[3]; s1[3] = 0;
	s2[0] = s[4]; s2[1] = s[5]; s2[2] = s[6]; s2[3] = 0;
	tryte high = s1, low = s2; 
	return tryte::word_to_int(high, low);
}

int floatval(char* s) {
	tryte high, low;
	float f;
	if(!sscanf(s, "%ff", &f)) {
		throw new error(WHERE, "Bad floating point value");
		return 0;
	}
	agdp::float2_to_float3(f, low, high);
	return tryte::word_to_int(low, high);
}
unsigned int strtoopc(char* s) {
	int i;
	for(i = -40; i <= 40; i++) {
		if(strcasecmp(s, machine::opcode_to_string(i)) == 0) {
			return i;
		}
	}

	string err = string("Couldn't identify opcode ") + s; 
	throw new error((WHERE), err);
}

/* Assembler class functions go here
 *
 *
 *
 *
 */

void assembler::org(int p) { pc = p; }
void assembler::dt(int v) { m.memref(pc++) = v; }
void assembler::dtstring(const char* s) {
	while(*s) {
		dt(asciitoternary(*s));
		s++;
	}
}
void assembler::dw(int v) { 
	tryte::int_to_word(v, m.memref(pc), m.memref(pc+1)); 
	pc += 2;
}
bool islocal(const string& s) { return s[0] == '.'; }

void assembler::equ(const char* c, int p) { 
	if(!islocal(c)) {
		variables[c] = p; 
		variable_tracker.insert(pair<string, string>(filename, c));
		return;
	} else {
		map<string,int>::iterator closest = labels.end();
		
		for(map<string,int>::iterator i = labels.begin();
				i != labels.end(); i++) {
			if((*i).first.find(".", 0) < (*i).first.npos) continue;
			if((*i).second <= pc) {
				if(closest == labels.end() ||
					((*closest).second < (*i).second)) {
					closest = i;
				}
			}
		}

		if(closest != labels.end()) {
			string name = (*closest).first + string(c);
			equ(name.c_str(), p);
		} else throw new error(WHERE, string("Unable to add local variable ") + c);
	}
}

void assembler::ldef(const char* c) { 
	if(!islocal(c)) {
		if(labels.find(c) != labels.end()) {
			if(state == INITIAL_SWEEP) printf("Warning! Redefining '%s'\n", c);
		}
		labels[c] = pc; 
		label_tracker.insert(pair<string, string>(filename, c));

		return;
	} else {
		map<string,int>::iterator closest = labels.end();
		
		for(map<string,int>::iterator i = labels.begin();
				i != labels.end(); i++) {
			if((*i).first.find(".", 0) < (*i).first.npos) continue;
			if((*i).second <= pc) {
				if(closest == labels.end() ||
					((*closest).second < (*i).second)) {
					closest = i;
				}
			}
		}

		if(closest != labels.end()) {
			string name = (*closest).first + string(c);
			ldef(name.c_str());
		} else throw new error(WHERE, string("Unable to add local variable ") + c);
	}
}

int assembler::label_eval(const char* c) { 
	if(islocal(c)) {
		map<string,int>::iterator closest = labels.end();
		
		for(map<string,int>::iterator i = labels.begin();
				i != labels.end(); i++) {
			if((*i).first.find(".", 0) < (*i).first.npos) continue;
			if((*i).second <= pc) {
				if(closest == labels.end() ||
						((*closest).second <
						 (*i).second)) {
					closest = i;
				}
			}
		}

		if(closest != labels.end()) {
			string name = (*closest).first + string(c);
			return label_eval(name.c_str());
		} else {
			string err = string("Unable to evaluate local variable ") + c;
			throw new error(WHERE, err);
		}
	}
	if(labels.find(c) != labels.end()) {
		return labels[c]; 
	} else if(variables.find(c) != variables.end()) {
		return variables[c];
	}


	if(state == FINAL_SWEEP) {
		string err = string("Unresolved varaible ") + c;
		throw new error(WHERE, err);
	}
	else return 0; // It's only the initial sweep
}

void assembler::addop(char* c, op_mode mode, int val) {
	unsigned int opc = strtoopc(c);

	//if(state == FINAL_SWEEP) printf("%d: %s %d %d\n", pc, c, mode, val);
	switch(mode) {
		case OP_IMPLICIT: 
			m.memref(pc++) = m.qop(machine::IMPLICIT, opc); break;
		case OP_ACC: 
			m.memref(pc++) = m.qop(machine::ACC, opc); break;
		case OP_XY: 
			m.memref(pc++) = m.qop(machine::XY, opc); break;
		case OP_IMMEDIATE: 
			m.memref(pc++) = m.qop(machine::IMMEDIATE, opc); break;
		case OP_ABSOLUTE: 
			m.memref(pc++) = m.qop(machine::ABS, opc); break;
		case OP_ABSOLUTE_X: 
			m.memref(pc++) = m.qop(machine::AX, opc); break;
		case OP_ABSOLUTE_Y: 
			m.memref(pc++) = m.qop(machine::AY, opc); break;
		case OP_INDIRECT: 
			m.memref(pc++) = m.qop(machine::INDIRECT, opc); break;
		case OP_INDIRECT_X: 
			m.memref(pc++) = m.qop(machine::INDX, opc); break;
		case OP_INDIRECT_Y: 
			m.memref(pc++) = m.qop(machine::INDY, opc); break;
	}
	if(mode == OP_IMMEDIATE) { m.memref(pc++) = val; }
	else if(mode > OP_IMMEDIATE) { 
		tryte::int_to_word(val, m.memref(pc), m.memref(pc+1)); 
		pc += 2;
	}
}

void help(char* argv0) {
	printf("\nUsage: %s [-h] [-v] [-o outfile] file [file2 ... filen]\n"
		"\t-h\tThis message\n"
		"\t-v\tVerbose mode\n"
		"\t-o\tSpecify output file\n"
		"\n", argv0);
}

void assembler::verbose_info() {
	printf("LABELS:\n");
	printf("%40s\t%7s\t%7s\n", "LABEL", "DEC", "NON");
	multimap<string, string>::iterator prev = label_tracker.end();

	for(multimap<string, string>::iterator i = label_tracker.begin();
			i != label_tracker.end(); i++) {
		tryte high, low;
		int v = labels[(*i).second];
		tryte::int_to_word(v, high, low);
		if(prev == label_tracker.end() || 
				(*prev).first != (*i).first) {
			printf("%s:\n", (*i).first.c_str());
		}

		printf("%40s\t%7.6d\t%.3X:%.3X\n", (*i).second.c_str(),
			       	labels[(*i).second],
				high.nonaryhex(), low.nonaryhex());
		prev = i;
		
	} 

	printf("VARIABLES:\n");
	printf("%40s\t%7s\t%7s\n", "LABEL", "DEC", "NON");

	prev = variable_tracker.end();
	for(multimap<string, string>::iterator i = variable_tracker.begin();
			i != variable_tracker.end(); i++) {
		tryte high, low; 
		tryte::int_to_word(variables[(*i).second], high, low);
		
		if(prev == variable_tracker.end() || 
				(*prev).first != (*i).first) {
			printf("%s:\n", (*i).first.c_str());
		} 

		printf("%40s\t%7.6d\t%.3X:%.3X\n", (*i).second.c_str(), variables[(*i).second],
				high.nonaryhex(), low.nonaryhex());
		prev = i; 
		
	} 
}

void assembler::read_files(yyFlexLexer* lexer, int argc, int findex, char** argv) {
	for(int i = findex; i < argc; i++) {
		if(strcmp(argv[i], "stdin") == 0) {
			printf("Sorry, no stdin, need to scan the file twice :-(\n");
			exit(EXIT_FAILURE);
		}

		filename = argv[i];
		std::ifstream* input = new ifstream(filename);
		if(input->fail()) {
			throw new error(WHERE, string("Unable to open ") + filename + string(" for reading"));
		}
		lexer->switch_streams(input, NULL);
		line = 1;
		try {
			yyparse();
		} catch (error* e) {
			ostringstream msgstr;
			msgstr << "Error during assembly of " << filename << ", line " << lineno();
			throw new error(WHERE, msgstr.str(), e);
			e->trace();
			printf("\n");
			exit(EXIT_FAILURE);
		}
		delete input;
	}


}

yyFlexLexer* lexer = NULL;
assembler* assembler_instance = NULL;

void yylex() { lexer->yylex(); }
int assembler::lineno() { return line; }

int main(int argc, char* argv[]) {
	bool verbose = false;
	int optret;
	char* outfile = NULL;

	/* Check command arguments */
	while((optret = getopt(argc, argv, "o:hv")) != -1) {
		switch(optret) {
			case 'h': help(argv[0]); exit(EXIT_SUCCESS);
			case 'o': outfile = strdup(optarg); break;
			case 'v': verbose = true; break;
			case '?':
			case ':': help(argv[0]); exit(EXIT_FAILURE);
		}
	}
	if(outfile == NULL) outfile = "out.ternobj";
	int fileindex = optind;

	/* Initiate lexer and assembler */
	assembler_instance = new assembler();
	lexer = new yyFlexLexer();

	/* Initiation done */

	if(fileindex >= argc) {
		printf("No files to process\n");
		help(argv[0]);
		return EXIT_FAILURE;
	}
	
	try {
		/* Do the first assembly sweep */
		if(verbose) printf("Initial sweep...\n");
		assembler_instance->set_state(INITIAL_SWEEP);
		assembler_instance->set_pc(0);
		assembler_instance->read_files(lexer, argc, fileindex, argv);
	} catch (error* e) {
		printf("\nCaught error during assembly (first sweep) \n\n");
		e->trace();
		printf("\n");
		return EXIT_FAILURE;
	}

	/* Print info about labels and variables if necessary */
	if(verbose) { assembler_instance->verbose_info(); }

	try {
		/* Do the final assembly sweep */
		if(verbose) printf("Final sweep...\n");
		assembler_instance->set_state(FINAL_SWEEP);
		assembler_instance->set_pc(0);
		assembler_instance->read_files(lexer, argc, fileindex, argv);
	} catch (error* e) {
		printf("\nCaught error during assembly (final sweep) \n\n");
		e->trace();
		printf("\n");
		return EXIT_FAILURE;
	}
	/* Save to disk */
	if(verbose) printf("Done.\n");

	assembler_instance->save(outfile);

	if(verbose) printf("Saved to '%s'.\n", outfile);

	return EXIT_SUCCESS;
}
